# Copyright 2020 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np
from pyqubo import Binary, Spin, Array, Num


class TestArray(unittest.TestCase):

    def test_array_create_binary(self):
        array = Array.create('x', shape=(3, 3, 3), vartype='BINARY')
        self.assertTrue(array.shape == (3, 3, 3))
        self.assertTrue(array[0][0][0] == Binary('x[0][0][0]'))

    def test_array_create_spin(self):
        array = Array.create('x', shape=(3, 3, 3), vartype='SPIN')
        self.assertTrue(array.shape == (3, 3, 3))
        self.assertTrue(array[0][0][0] == Spin('x[0][0][0]'))

    def test_array_from_list(self):
        array = Array([[Binary('x0'), Binary('x1')], [Binary('x2'), Binary('x3')]])
        self.assertTrue(array.shape == (2, 2))
        self.assertTrue(array[0, 0] == Binary('x0'))
        self.assertTrue(array[:, 0] == Array([Binary('x0'), Binary('x2')]))
        self.assertTrue(array[[0, 1], 0] == Array([Binary('x0'), Binary('x2')]))
        self.assertRaises(TypeError, lambda: Array(1))
        self.assertTrue(Array([[1, 2], [3, 4]]) == Array([Array([1, 2]), Array([3, 4])]))
        self.assertTrue(Array([[1, 2], [3, 4]]) == Array([np.array([1, 2]), np.array([3, 4])]))

    def test_array_error(self):
        # raise ValueError when the size of the sub lists is not same
        with self.assertRaises(ValueError):
            Array([[Binary('x0'), Binary('x1')], [Binary('x2')]])

        # raise TypeError when index is neither int nor tuple[int]
        with self.assertRaises(TypeError):
            array = Array([Binary('x0'), Binary('x1')])
            array[1.5]

    def test_array_equality(self):
        array1 = Array([[Binary('x0'), Binary('x1')], [Binary('x2'), Binary('x3')]])
        array2 = Array([[Binary('x0'), Binary('x1')], [Binary('x2'), Binary('x3')]])
        array3 = Array([[Binary('x0'), Binary('x1')], [Binary('x2'), Binary('x4')]])
        self.assertTrue(array1 == array2)
        self.assertTrue(array1 != array3)
        self.assertTrue(not array1 == Binary('x2'))

    def test_array_iter(self):
        array = Array.create('x', shape=(2, 2), vartype='BINARY')
        self.assertTrue(len(array) == 2)
        self.assertTrue([e for e in array] == [array[0], array[1]])

    def test_array_from_numpy(self):
        array = Array(np.array([[1, 2], [3, 4]]))
        self.assertTrue(array == Array([[1, 2], [3, 4]]))

    def test_array_transpose(self):
        array = Array.create('x', (2, 3), 'BINARY')
        expected = Array([[Binary('x[0][0]'), Binary('x[1][0]')],
                          [Binary('x[0][1]'), Binary('x[1][1]')],
                          [Binary('x[0][2]'), Binary('x[1][2]')]])
        self.assertTrue(array.T == expected)

    def test_array_multiply(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        array2 = Array.create('y', (2, 2), 'BINARY')
        expected = Array([[Binary('x[0][0]') * Binary('y[0][0]'), Binary('x[0][1]') * Binary('y[0][1]')],
                          [Binary('x[1][0]') * Binary('y[1][0]'), Binary('x[1][1]') * Binary('y[1][1]')]])
        self.assertTrue(array1 * array2 == expected)

        expected2 = Array([[Binary('x[0][0]') * 2, Binary('x[0][1]') * 2],
                           [Binary('x[1][0]') * 2, Binary('x[1][1]') * 2]])
        self.assertTrue(array1 * 2 == expected2)
        self.assertTrue(2 * array1 == expected2)
        self.assertTrue(array1 * (2*np.ones((2, 2))) == expected2)

    def test_array_division(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        expected = Array([[Binary('x[0][0]') / 2.0, Binary('x[0][1]') / 2.0],
                          [Binary('x[1][0]') / 2.0, Binary('x[1][1]') / 2.0]])
        self.assertTrue(array1 / 2.0 == expected)

        self.assertRaises(ValueError, lambda: 1 / array1)
        self.assertRaises(ValueError, lambda: array1 / array1)

    def test_array_reshape(self):
        array = Array.create('a', shape=(2, 3), vartype='BINARY')
        reshaped = array.reshape((3, 2))
        self.assertTrue(reshaped.shape == (3, 2))
        expected = Array([[Binary('a[0][0]'), Binary('a[0][1]')],
                          [Binary('a[0][2]'), Binary('a[1][0]')],
                          [Binary('a[1][1]'), Binary('a[1][2]')]])
        self.assertTrue(reshaped == expected)


if __name__ == '__main__':
    unittest.main()


# Copyright 2020 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import Binary, Placeholder, Array, SubH, Constraint
from pyqubo import assert_qubo_equal
import numpy as np
import dimod

class TestModel(unittest.TestCase):

    def test_to_qubo(self):
        a, b = Binary("a"), Binary("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        qubo, offset = model.to_qubo()
        # assert_qubo_equal(qubo, {("a", "a"): 1.0, ("a", "b"): 1.0, ("b", "b"): 0.0})
        assert_qubo_equal(qubo, {("a", "a"): 1.0, ("a", "b"): 1.0})
        self.assertTrue(offset == -1)

    def test_to_qubo_with_index(self):
        a, b = Binary("a"), Binary("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        qubo, offset = model.to_qubo(index_label=True)
        # assert_qubo_equal(qubo, {(0, 0): 1.0, (0, 1): 1.0, (1, 1): 0.0})
        assert_qubo_equal(qubo, {(0, 0): 1.0, (0, 1): 1.0})
        self.assertTrue(offset == -1)

    def test_to_ising(self):
        a, b = Binary("a"), Binary("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        linear, quad, offset = model.to_ising()
        self.assertTrue(linear == {'a': 0.75, 'b': 0.25})
        assert_qubo_equal(quad, {('a', 'b'): 0.25})
        self.assertTrue(offset == -0.25)

    def test_to_ising_with_index(self):
        a, b = Binary("a"), Binary("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        linear, quad, offset = model.to_ising(index_label=True)
        self.assertTrue(linear == {0: 0.75, 1: 0.25})
        assert_qubo_equal(quad, {(0, 1): 0.25})
        self.assertTrue(offset == -0.25)

    def test_decode_sample(self):
        x = Array.create("x", (2, 2), vartype="BINARY")
        exp = SubH((x[1, 1] + x[0, 1] + x[0, 0] - 1) ** 2, label="const")
        model = exp.compile()

        # type of solution is dict[label, int]

        ## vartype = BINARY
        sample = {'x[0][1]': 1, 'x[1][1]': 1, 'x[0][0]': 0}
        decoded_sample = model.decode_sample(sample, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 0)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 1)

        sample = {'x[0][1]': 1, 'x[1][1]': 0, 'x[0][0]': 0}
        decoded_sample = model.decode_sample(sample, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 0)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 0)

        ## vartype = SPIN
        sample = {'x[0][1]': 1, 'x[1][1]': 1, 'x[0][0]': -1}
        decoded_sample = model.decode_sample(sample, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 1)

        sample = {'x[0][1]': 1, 'x[1][1]': -1, 'x[0][0]': -1}
        decoded_sample = model.decode_sample(sample, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == -1)

        # type of solution is list[int]

        ## vartype = BINARY
        sample = {'x[0][1]': 1, 'x[1][1]': 0, 'x[0][0]': 1}
        list_sample = [sample[v] for v in model.variables]
        decoded_sample = model.decode_sample(list_sample, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 0)

        sample = {'x[0][1]': 1, 'x[1][1]': 0, 'x[0][0]': 0}
        list_sample = [sample[v] for v in model.variables]
        decoded_sample = model.decode_sample(list_sample, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 0)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 0)

        ## vartype = SPIN
        sample = {'x[0][1]': 1, 'x[1][1]': 1, 'x[0][0]': -1}
        list_sample = [sample[v] for v in model.variables]
        decoded_sample = model.decode_sample(list_sample, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 1)

        sample = {'x[0][1]': 1, 'x[1][1]': -1, 'x[0][0]': -1}
        list_sample = [sample[v] for v in model.variables]
        decoded_sample = model.decode_sample(list_sample, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == -1)

        # type of solution is dict[index_label(int), bit]
        ## vartype = BINARY
        sample = {'x[0][1]': 1, 'x[1][1]': 0, 'x[0][0]': 1}
        sample_index = {i: sample[v] for i, v in enumerate(model.variables)}
        decoded_sample = model.decode_sample(sample_index, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 0)

        sample = {'x[0][1]': 1, 'x[1][1]': 0, 'x[0][0]': 0}
        sample_index = {v: sample[v] for v in model.variables}
        decoded_sample = model.decode_sample(sample_index, vartype="BINARY")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == 0)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 0)

        ## vartype = SPIN
        sample = {'x[0][1]': 1, 'x[1][1]': 1, 'x[0][0]': -1}
        sample_index = {v: sample[v] for v in model.variables}
        decoded_sample = model.decode_sample(sample_index, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 1)
        self.assertTrue(decoded_sample.energy == 1)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == 1)

        sample = {'x[0][1]': 1, 'x[1][1]': -1, 'x[0][0]': -1}
        sample_index = {v: sample[v] for v in model.variables}
        decoded_sample = model.decode_sample(sample_index, vartype="SPIN")
        self.assertTrue(decoded_sample.sample == sample)
        self.assertTrue(len([label for label, energy in decoded_sample.subh.items() if energy > 0]) == 0)
        self.assertTrue(decoded_sample.energy == 0)
        self.assertTrue(decoded_sample.array("x", (0, 0)) == -1)
        self.assertTrue(decoded_sample.array("x", (0, 1)) == 1)
        self.assertTrue(decoded_sample.array("x", (1, 1)) == -1)

        # 異常系は普通に異常終了するので、無視で。

        # invalid solution
        # self.assertRaises(ValueError, lambda: model.decode_sample([1, 1, 1, 1], vartype="BINARY"))
        # qubo = model.to_qubo()
        # self.assertRaises(ValueError, lambda: model.decode_sample(
        #     {'x[0][2]': 1, 'x[1][1]': 0, 'x[0][0]': 1}, vartype="BINARY"))
        # self.assertRaises(TypeError, lambda: model.decode_sample((1, 1), vartype="BINARY"))

    def test_work_with_dimod(self):
        S = Array.create('S', 3, "SPIN")
        H = 0.8 * S[0] * S[1] + S[1] * S[2] + 1.1 * S[2] * S[0] + 0.5 * S[0]
        model = H.compile()

        # with index_label=False
        binary_bqm = model.to_bqm(index_label=False)
        sampler = dimod.ExactSolver()
        sampleset = sampler.sample(binary_bqm)
        decoded_samples = model.decode_sampleset(sampleset)
        best_sample = min(decoded_samples, key=lambda s: s.energy)
        self.assertTrue(best_sample.array("S", 0) == 0)
        self.assertTrue(best_sample.array("S", 1) == 0)
        self.assertTrue(best_sample.array("S", 2) == 1)
        self.assertTrue(np.isclose(best_sample.energy, -1.8))

        # with index_label=True
        binary_bqm = model.to_bqm(index_label=True)
        sampler = dimod.ExactSolver()
        sampleset = sampler.sample(binary_bqm)
        decoded_samples = model.decode_sampleset(sampleset)
        best_sample = min(decoded_samples, key=lambda s: s.energy)
        self.assertTrue(best_sample.array("S", 0) == 0)
        self.assertTrue(best_sample.array("S", 1) == 0)
        self.assertTrue(best_sample.array("S", 2) == 1)
        self.assertTrue(np.isclose(best_sample.energy, -1.8))


    def test_placeholders(self):

        S = Array.create('S', 3, "SPIN")
        p1, p2 = Placeholder("p1"), Placeholder("p2")
        H = p1 * S[0] * S[1] + S[1] * S[2] + p2 * S[2] * S[0] + 0.5 * S[0]
        model = H.compile()
        feed_dict = {"p1": 0.8, "p2": 1.1}
        binary_bqm = model.to_bqm(feed_dict=feed_dict)
        sampler = dimod.ExactSolver()
        sampleset = sampler.sample(binary_bqm)
        decoded_samples = model.decode_sampleset(sampleset, feed_dict=feed_dict)
        best_sample = min(decoded_samples, key=lambda s: s.energy)

        self.assertTrue(best_sample.array("S", 0) == 0)
        self.assertTrue(best_sample.array("S", 1) == 0)
        self.assertTrue(best_sample.array("S", 2) == 1)
        self.assertTrue(np.isclose(best_sample.energy, -1.8))

    def test_compiler_strength(self):
        a, b, c, d = Binary("a"), Binary("b"), Binary("c"), Binary("d")
        H = a * b * c - 2 * c * d - 3 * a * b
        model = H.compile(strength=Placeholder("s"))
        feed_dict = {"s": 10.0}
        qubo, offset = model.to_qubo(feed_dict=feed_dict)
        sampler = dimod.ExactSolver()
        sampleset = sampler.sample_qubo(qubo)
        decoded_samples = model.decode_sampleset(sampleset, feed_dict=feed_dict)
        best_sample = min(decoded_samples, key=lambda s: s.energy)
        
        self.assertTrue(qubo[('a', 'a * b')] == -20)
        self.assertTrue(best_sample.value(b) == 1)
        self.assertTrue(best_sample.value(a) == 1)
        self.assertTrue(best_sample.value(b) == 1)
        self.assertTrue(best_sample.value(c) == 1)
        self.assertTrue(best_sample.value(d) == 1)
        self.assertTrue(np.isclose(best_sample.energy, -4))

    def test_constraint(self):
        sampler = dimod.ExactSolver()
        x = Array.create('x',shape=(3),vartype="BINARY")
        H = Constraint(x[0]*x[1]*x[2],label="C1")
        model = H.compile()
        bqm = model.to_bqm()
        responses = sampler.sample(bqm)
        solutions = model.decode_sampleset(responses)

        for sol in solutions:
            if sol.energy==1.0:
                self.assertEqual(sol.subh['C1'], 1.0)


    def test_higher_order(self):
        x = Array.create('x', 5, 'BINARY')
        exp = x[0]*x[1]*x[2]*x[3]
        model = exp.compile(strength=10)

        # sample = {'x[0]': 1, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, '0*1': 1, '2*3': 1}
        sample = {'x[0]': 1, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, 'x[0] * x[1]': 1, 'x[2] * x[3]': 1}
        e = model.energy(sample, vartype='BINARY')
        self.assertEqual(e, 1.0)

        # sample = {'x[0]': 0, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, '0*1': 0, '2*3': 1}
        sample = {'x[0]': 0, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, 'x[0] * x[1]': 0, 'x[2] * x[3]': 1}
        e = model.energy(sample, vartype='BINARY')
        self.assertEqual(e, 0.0)

        # sample = {'x[0]': 1, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, '0*1': 0, '2*3': 1}
        sample = {'x[0]': 1, 'x[1]': 1, 'x[2]': 1, 'x[3]': 1, 'x[0] * x[1]': 0, 'x[2] * x[3]': 1}
        e = model.energy(sample, vartype='BINARY')
        self.assertEqual(e, 10.0)



if __name__ == '__main__':
    unittest.main()



import unittest
from pyqubo import Binary, Spin, WithPenalty, SubH, Constraint, assert_qubo_equal, Placeholder


class TestExpress(unittest.TestCase):

    def compile_check(self, exp, expected_qubo, expected_offset, feed_dict={}):
        model = exp.compile(strength=5)
        qubo, offset = model.to_qubo(feed_dict=feed_dict)
        assert_qubo_equal(qubo, expected_qubo)
        self.assertEqual(offset, expected_offset)

    def test_compile_binary2(self):
        a, b = Binary("a"), Binary("b")
        exp =  a + 2*a + b -1
        # expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 1.0, ('b', 'b'): 0.0}
        expected_qubo = {('a', 'a'): 3.0, ('b', 'b'): 1.0}
        expected_offset = -1
        self.compile_check(exp, expected_qubo, expected_offset)

    def test_compile_binary(self):
        a, b = Binary("a"), Binary("b")
        exp = 1 + a*b + a - 2
        # expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 1.0, ('b', 'b'): 0.0}
        expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 1.0}
        expected_offset = -1
        self.compile_check(exp, expected_qubo, expected_offset)

    def test_compile_spin(self):
        exp = 2*(Spin("a") - Binary("b"))
        expected_qubo = {('a', 'a'): 4.0, ('b', 'b'): -2.0}
        expected_offset = -2.0
        self.compile_check(exp, expected_qubo, expected_offset)
    

    def test_compile_expand_add(self):
        a, b = Binary("a"), Binary("b")
        exp = (a+b)*(a-b)
        # expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 0.0, ('b', 'b'): -1.0}
        expected_qubo = {('a', 'a'): 1.0, ('b', 'b'): -1.0}
        expected_offset = 0.0
        self.compile_check(exp, expected_qubo, expected_offset)
    
    
    def test_compile_div(self):
        a, b = Binary("a"), Binary("b")
        exp = a*b / 2 + 1
        # expected_qubo = {('a', 'a'): 0.0, ('a', 'b'): 0.5, ('b', 'b'): 0.0}
        expected_qubo = {('a', 'b'): 0.5}
        expected_offset = 1.0
        q, offset = exp.compile().to_qubo()
        self.compile_check(exp, expected_qubo, expected_offset)
    
    def test_compile_power(self):
        a, b = Binary("a"), Binary("b")
        exp = (a+b)**3
        expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 6.0, ('b', 'b'): 1.0}
        expected_offset = 0.0
        q, offset = exp.compile().to_qubo()
        self.compile_check(exp, expected_qubo, expected_offset)
    
    def test_compile_neg(self):
        exp = -Binary("a")
        expected_qubo = {('a', 'a'): -1.0}
        expected_offset = 0.0
        self.compile_check(exp, expected_qubo, expected_offset)
    
    def test_compile_placeholder(self):
        a, b = Binary("a"), Binary("b")
        p, q, r = Placeholder("p"), Placeholder("q"), Placeholder("r")
        exp = r*(q*p*(a+b)**2 + q)
        expected_qubo = {('a', 'a'): 12.0, ('a', 'b'): 24.0, ('b', 'b'): 12.0}
        expected_offset = 4
        feed_dict={"p": 3, "q": 2, "r": 2}
        q, offset = exp.compile().to_qubo(feed_dict=feed_dict)
        self.compile_check(exp, expected_qubo, expected_offset, feed_dict)
    
    def test_compile_subh(self):
        a, b = Binary("a"), Binary("b")
        p = Placeholder("p")
        exp = p * SubH((a+b-1)**2, label="subh") + a*b
        expected_qubo = {('a', 'a'): -3.0, ('a', 'b'): 7.0, ('b', 'b'): -3.0}
        expected_offset = 3
        feed_dict={"p": 3}
        self.compile_check(exp, expected_qubo, expected_offset, feed_dict)
    
    def test_compile_constraint(self):
        a, b, c = Binary("a"), Binary("b"), Binary("c")
        exp = Constraint(a*b*c, label="constraint")
        #exp=a*b*c
        # expected_qubo = {('a', '0*1'): -10.0, ('b', '0*1'): -10.0, ('0*1', '0*1'): 15.0, ('a', 'a'): 0.0, ('a', 'b'): 5.0, ('c', '0*1'): 1.0, ('b', 'b'): 0.0, ('c', 'c'): 0.0}
        model = exp.compile()
        expected_qubo = {('a', 'a * b'): -10.0, ('b', 'a * b'): -10.0, ('a * b', 'a * b'): 15.0, ('a', 'b'): 5.0, ('c', 'a * b'): 1.0}
        expected_offset = 0
        self.compile_check(exp, expected_qubo, expected_offset, feed_dict={})

    
    def test_compile_with_penalty(self):
        class CustomPenalty(WithPenalty):
            def __init__(self, hamiltonian, penalty, label):
                super().__init__(hamiltonian, penalty, label)

        a, b = Binary("a"), Binary("b")
        p = Placeholder("p")
        custom_penalty = p * CustomPenalty(a+b, a*b, "label")
        expected_qubo = {('a', 'a'): 2.0, ('a', 'b'): 1.0, ('b', 'b'): 2.0}
        expected_offset = 0.0
        feed_dict={"p": 2}
        self.compile_check(custom_penalty, expected_qubo, expected_offset, feed_dict)
    

if __name__ == '__main__':
    unittest.main()



# Copyright 2020 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyqubo import OneHotEncInteger, OrderEncInteger, Placeholder, LogEncInteger, UnaryEncInteger
import dimod
from pyqubo import assert_qubo_equal


class TestInteger(unittest.TestCase):

    def test_one_hot_enc_integer(self):
        a = OneHotEncInteger("a", (0, 4), strength=Placeholder("s"))
        H = (a - 3) ** 2
        model = H.compile()
        q, offset = model.to_qubo(feed_dict={"s": 10.0})
        sampleset = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(
            sampleset, feed_dict={"s": 10.0})
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.value(a) == 3)

        # will raise runtime error if the constraint is broken, when evaluating the value of the one-hot-integer object.
        worst = max(decoded, key=lambda x: x.energy)
        self.assertRaises(RuntimeError, lambda: worst.value(a))

        self.assertTrue(a.value_range == (0, 4))

        # expected_q = {('a[0]', 'a[1]'): 20.0,
        #      ('a[0]', 'a[2]'): 20.0,
        #      ('a[0]', 'a[3]'): 20.0,
        #      ('a[0]', 'a[4]'): 20.0,
        #      ('a[1]', 'a[2]'): 24.0,
        #      ('a[1]', 'a[3]'): 26.0,
        #      ('a[1]', 'a[4]'): 28.0,
        #      ('a[2]', 'a[3]'): 32.0,
        #      ('a[2]', 'a[4]'): 36.0,
        #      ('a[3]', 'a[4]'): 44.0,
        #      ('a[0]', 'a[0]'): -10.0,
        #      ('a[1]', 'a[1]'): -15.0,
        #      ('a[2]', 'a[2]'): -18.0,
        #      ('a[3]', 'a[3]'): -19.0,
        #      ('a[4]', 'a[4]'): -18.0}
        # expected_offset = 19
        # assert_qubo_equal(q, expected_q)
        # self.assertTrue(offset == expected_offset)
  
    def test_one_hot_enc_integer_equal(self):
        a = OneHotEncInteger("a", (0, 4), strength=Placeholder("s"))
        b = OneHotEncInteger("b", (0, 4), strength=Placeholder("s"))
        M = 2.0
        H = (a + b - 5) ** 2 + M * (a.equal_to(3) - 1)**2
        model = H.compile()
        q, offset = model.to_qubo(feed_dict={"s": 10.0})
        sampleset = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(
            sampleset, feed_dict={"s": 10.0})
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.value(a) == 3)
        self.assertTrue(best.value(b) == 2)
        self.assertTrue(best.subh["a_const"]==0)
        self.assertTrue(best.subh["b_const"]==0)
        self.assertEqual(len(best.constraints(only_broken=True)), 0)

    def test_order_enc_integer(self):
        a = OrderEncInteger("a", (0, 4), strength=Placeholder("s"))
        model = ((a - 3) ** 2).compile()
        q, offset = model.to_qubo(feed_dict={"s": 10.0})
        # expected_q = {
        #     ('a[3]', 'a[2]'): -8.0,
        #     ('a[0]', 'a[1]'): -8.0,
        #     ('a[3]', 'a[0]'): 2.0,
        #     ('a[2]', 'a[0]'): 2.0,
        #     ('a[1]', 'a[1]'): 5.0,
        #     ('a[3]', 'a[1]'): 2.0,
        #     ('a[2]', 'a[1]'): -8.0,
        #     ('a[3]', 'a[3]'): 5.0,
        #     ('a[0]', 'a[0]'): -5.0,
        #     ('a[2]', 'a[2]'): 5.0
        # }
        response = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(
            response, feed_dict={"s": 10.0})
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.subh["a"]==3)
        self.assertTrue(a.value_range == (0, 4))
        # assert_qubo_equal(q, expected_q)

    def test_order_enc_integer_more_than(self):
        a = OrderEncInteger("a", (0, 4), strength=5.0)
        b = OrderEncInteger("b", (0, 4), strength=5.0)
        model = ((a - b) ** 2 + (1 - a.more_than(1)) ** 2 + (1 - b.less_than(3)) ** 2).compile()
        q, offset = model.to_qubo()
        sampleset = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(sampleset)
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.subh["a"]==2)
        self.assertTrue(best.subh["b"]==2)

    def test_log_enc_integer(self):
        a = LogEncInteger("a", (0, 4))
        b = LogEncInteger("b", (0, 4))
        M = 2.0
        H = (2 * a - b - 1) ** 2 + M * (a + b - 5) ** 2
        model = H.compile()
        q, offset = model.to_qubo()
        sampleset = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(sampleset)
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.value(a) == 2)
        self.assertTrue(best.value(b) == 3)
        self.assertTrue(a.value_range == (0, 4))
        self.assertTrue(b.value_range == (0, 4))


    def test_unary_enc_integer(self):
        a = UnaryEncInteger("a", (0, 3))
        b = UnaryEncInteger("b", (0, 3))
        M = 2.0
        H = (2 * a - b - 1) ** 2 + M * (a + b - 3) ** 2
        model = H.compile()
        q, offset = model.to_qubo()
        sampleset = dimod.ExactSolver().sample_qubo(q)
        decoded = model.decode_sampleset(sampleset)
        best = min(decoded, key=lambda x: x.energy)
        self.assertTrue(best.value(a) == 1)
        self.assertTrue(best.value(b) == 2)
        self.assertTrue(a.value_range == (0, 3))
        self.assertTrue(b.value_range == (0, 3))

        # expected_q = {('a[0]', 'a[1]'): 12.0,
        #      ('a[0]', 'a[2]'): 12.0,
        #      ('a[0]', 'b[0]'): 0.0,
        #      ('a[0]', 'b[1]'): 0.0,
        #      ('a[0]', 'b[2]'): 0.0,
        #      ('a[1]', 'a[2]'): 12.0,
        #      ('a[1]', 'b[0]'): 0.0,
        #      ('a[1]', 'b[1]'): 0.0,
        #      ('a[1]', 'b[2]'): 0.0,
        #      ('a[2]', 'b[0]'): 0.0,
        #      ('a[2]', 'b[1]'): 0.0,
        #      ('a[2]', 'b[2]'): 0.0,
        #      ('b[0]', 'b[1]'): 6.0,
        #      ('b[0]', 'b[2]'): 6.0,
        #      ('b[1]', 'b[2]'): 6.0,
        #      ('a[0]', 'a[0]'): -10.0,
        #      ('a[1]', 'a[1]'): -10.0,
        #      ('a[2]', 'a[2]'): -10.0,
        #      ('b[0]', 'b[0]'): -7.0,
        #      ('b[1]', 'b[1]'): -7.0,
        #      ('b[2]', 'b[2]'): -7.0}
        # assert_qubo_equal(q, expected_q)
    

if __name__ == '__main__':
    unittest.main()
